from ngsolve import *
import ngsolve as ng
from ngsolve import x, y, z, curl, Mesh, HCurl, BlockMatrix
from netgen.csg import OrthoBrick, unit_cube, CSGeometry, Pnt
from math import pi, log
from array import array
import math as m
import sympy as sm
import scipy.sparse as sp
import numpy as np
from ngsolve.la import EigenValues_Preconditioner
import time


box = OrthoBrick(Pnt(-1,-1,-1),Pnt(1,1,1)).bc("outer")
geo = CSGeometry()
geo.Add(box)

beta = 1 

X, Y, Z = sm.symbols('X Y Z')

ex = ((1 - X**2)*(1 - Y**2)*(1 - Z**2)*((X**2 + Y**2 + Z**2)**(0.6)),
      (1 - X**2)*(1 - Y**2)*(1 - Z**2)*((X**2 + Y**2 + Z**2)**(0.6)),
      (1 - X**2)*(1 - Y**2)*(1 - Z**2)*((X**2 + Y**2 + Z**2)**(0.6)))

def smcurl(M):
    return (sm.diff(M[2], Y) - sm.diff(M[1], Z),
            sm.diff(M[0], Z) - sm.diff(M[2], X),
            sm.diff(M[1], X) - sm.diff(M[0], Y))


def str2coef(symbol):
    print(symbol)
    return ng.CoefficientFunction(eval(str(symbol)
                                       .replace("X", "x")
                                       .replace("Y", "y")
                                       .replace("Z", "z")))

exact = str2coef(ex)
curlex = smcurl(ex)
curl_exact = str2coef(curlex)
curlcurl_exact = str2coef(smcurl(smcurl(ex)))
src = curlcurl_exact +beta*exact

P_order = 1
Int_mesh_size = 2 

L2_u_err_list = array('d', [0])
HCurl_u_err_list = array('d', [0])
Mesh_size_list = array('d', [0])


ngmesh = geo.GenerateMesh(maxh=Int_mesh_size)
mesh = Mesh(ngmesh)

V = HCurl(mesh, order=P_order, type1=True,dirichlet="outer" )

# define trial- and test-functions
u,v = V.TnT()

# the bilinear-form
a = ng.BilinearForm(V)
a += ng.SymbolicBFI(curl(u) * curl(v)+beta*u*v)

# the right hand side
fv = ng.LinearForm(V)
fv += ng.SymbolicLFI(src*v)

# the solution field
gfv = ng.GridFunction(V)

ng.Draw(gfv, mesh, "gfv")

def SolveBVP():
    gfv.Update()
    e.Update()
    a.Assemble()
    fv.Assemble()
    gfv.vec.data = a.mat.Inverse(V.FreeDofs()) * fv.vec

    #Redraw (blocking=True)
    ng.Draw(gfv, mesh, "gfv")
    
def CalcError():
    curl_u_gfv = curl(gfv)

    #U
    L2_u_err = ng.sqrt(ng.Integrate((gfv-exact)*(gfv-exact), mesh))
    HCurl_u_err = ng.sqrt(ng.Integrate(
        (gfv-exact)*(gfv-exact) +
        (curl_u_gfv-curl_exact)*(curl_u_gfv-curl_exact), mesh))
    
    L2_u_err_list.append(L2_u_err)
    HCurl_u_err_list.append(HCurl_u_err)

    #--------------------------------
    # Calc. local errors
    err = (gfv-exact)*(gfv-exact) + (curl_u_gfv-curl_exact)*(curl_u_gfv-curl_exact)
    elerr = ng.Integrate (err, mesh, VOL, element_wise=True)   
    maxerr = max(elerr)
    l.append ( (V.ndof, sqrt(sum(elerr)) ))
    print ("maxerr = ", maxerr)
    
    for el in mesh.Elements():
        mesh.SetRefinementFlag(el, elerr[el.nr] > 0.25*maxerr)
    #---------------------------------------------

    return elerr
cn=1
Mesh_size_list.append(V.ndof)
with TaskManager():
    while V.ndof < 5000:  
        SolveBVP()
        elerr = CalcError()
        print("----------------------Refine No =",cn, "----------------------")
        mesh.Refine()
        V.Update()
        cn=cn+1
        Mesh_size_list.append(V.ndof)
SolveBVP()
elerr = CalcError()
Ref_num=cn


# VTKOutput object
vtk = VTKOutput(ma=mesh,
                coefs=[gfv],
                names = ["gfv"],
                filename="Simple_Exampl_adaptive")
vtk.Do()
#=============================================================================
print("-------------L2 Norms---------------")
print("------------------------------------")
print("|", "  h  ", "|","L2 Err.(U)","|", "HCurl Err_U", "|")
print("------------------------------------")
for i in range(1, Ref_num+1):
    print("|", "%5d" % Mesh_size_list[i],
    "|", "%8.4E" % L2_u_err_list[i], "|", "%8.4E " % HCurl_u_err_list[i], "|")
print("------------------------------------")


